/*

  This is a part of Layercut.dll source code.
  
  Copyright (C) 2001-2002 Sylvain Rouquette
  rouque_s@epita.fr
  http://www.epita.fr:8000/~rouque_s

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#include "layercut.h"
#include "debug.h"
#include "safestr.h"

void PreMultiplyRGBChannels(HBITMAP &bmp, LPBYTE pBitmapBits)
{
	BITMAP bm;
	GetObject(bmp, sizeof(BITMAP), &bm);
	BYTE *pPixel = pBitmapBits;

	for (int i = 0; i < bm.bmHeight * bm.bmWidth; ++i)
	{
		pPixel[0] = pPixel[0] * pPixel[3] / 255;
		pPixel[1] = pPixel[1] * pPixel[3] / 255;
		pPixel[2] = pPixel[2] * pPixel[3] / 255;

		pPixel+= 4;
	}
}

/**************************************************************************************/

HRGN BitmapToRegionAlpha(HBITMAP hbm, BYTE clrTransp, int xoffset, int yoffset)
{
  // start with a completely transparent rgn
	// this is more correct as no bmp, should render a transparent background
  HRGN hRgn = CreateRectRgn(0,0,0,0);

  if (hbm)
	{
		// create a dc for the 32 bit dib
    HDC hdcMem = CreateCompatibleDC(NULL);
		if (hdcMem)
		{
			VOID *pbits32;
			HBITMAP hbm32;
			BITMAP bm;
			// get the size
			GetObject(hbm, sizeof(BITMAP), &bm);

			BITMAPINFOHEADER bmpInfo32;
			bmpInfo32.biSize					= sizeof(BITMAPINFOHEADER);
			bmpInfo32.biWidth					= bm.bmWidth;
			bmpInfo32.biHeight				= bm.bmHeight;
			bmpInfo32.biPlanes				= 1;
			bmpInfo32.biBitCount			= 32;
			bmpInfo32.biCompression		= BI_RGB;
			bmpInfo32.biSizeImage			= 0;
			bmpInfo32.biXPelsPerMeter	= 0;
			bmpInfo32.biYPelsPerMeter	= 0;
			bmpInfo32.biClrUsed				= 0;
			bmpInfo32.biClrImportant	= 0;

			hbm32 = CreateDIBSection(hdcMem, (BITMAPINFO *)&bmpInfo32, DIB_RGB_COLORS, &pbits32, NULL, 0);
			if (hbm32)
			{
				HBITMAP hbmOld32 = (HBITMAP)SelectObject(hdcMem, hbm32);

				// Create a DC just to copy the bitmap into the memory D
				HDC hdcTmp = CreateCompatibleDC(hdcMem);
				if (hdcTmp)
				{
					// Get how many bytes per row we have for the bitmap bits (rounded up to 32 bits
					int y = 0;
					BITMAP bm32;
					GetObject(hbm32, sizeof(bm32), &bm32);
					bm32.bmWidthBytes += bm32.bmWidthBytes % 4;

					// Copy the bitmap into the memory D
					HBITMAP hbmOld = (HBITMAP)SelectObject(hdcTmp, hbm);

					BitBlt(hdcMem,0,0,bm.bmWidth,bm.bmHeight,hdcTmp,0,0,SRCCOPY);

					// Scan each bitmap row from bottom to top (the bitmap is inverted vertically
					BYTE *p;
					BYTE *p32 = (BYTE *)bm32.bmBits + (bm32.bmHeight - 1) * bm32.bmWidthBytes + 3;
					while (y < bm.bmHeight)
					{
						int x = 0;
						while ( x < bm.bmWidth )
						{
							int x0 = 0;
							// loop through all transparent pixels...
							while ( x < bm.bmWidth )
							{
								if (p32[4 * x] > clrTransp)
									break;
								x++;
							}
							// set first non transparent pixel
							x0 = x;
							// loop through all non transparent pixels
							while ( x < bm.bmWidth )
							{
								if (p32[4 * x] <= clrTransp)
									break;
								x++;
							}
							// if found one or more non-transparent pixels in a row, add them to the rgn...
							if ((x-x0)>0)
							{
								HRGN hTempRgn = CreateRectRgn(x0 + xoffset, y + yoffset, x + xoffset, y+1 + yoffset);
								CombineRgn(hRgn, hRgn, hTempRgn, RGN_OR);
								DeleteObject(hTempRgn);
							}
							x++;
						}
						y++;
						p32 -= bm32.bmWidthBytes;
					}
					// Clean up
					SelectObject(hdcTmp, hbmOld);
					DeleteDC(hdcTmp);
				}
				SelectObject(hdcMem, hbmOld32);
				DeleteObject(hbm32);
			}
			DeleteDC(hdcMem);
		}
	}
	return hRgn;
}
/**************************************************************************************/

CImage::CImage(HBITMAP bmp, CAnimation *anim)
{

	image = bmp;

	if (anim->is_opaque)
		return;

	if (anim->has_alpha)
	{
		BITMAP bm;
		GetObject(bmp, sizeof(BITMAP), &bm);
		PreMultiplyRGBChannels(image, (LPBYTE)bm.bmBits);
		region = BitmapToRegionAlpha(image, alpha_threshold, 0, 0);
	}
	else
		region = BitmapToRegion(image, RGB(255, 0, 255), 0x101010, 0, 0);
}

/**************************************************************************************/

int CAnimation::init(string &n, CAnimation *img)
{
	HBITMAP bmp;

	bmp = LoadLSImage(n.c_str(), NULL);
	if (bmp)
	{
		BITMAP bm;
		GetObject(bmp, sizeof(BITMAP), &bm);
		if (!(img->width || img->height))
		{
			img->width = bm.bmWidth;
			img->height = bm.bmHeight;
		}
		name = n;
		speed = img->speed;
		loop = img->loop;
		finish = img->finish;
		priority = img->priority;
		is_opaque = img->is_opaque;
		min = img->min;
		max = img->max;
		computeImage(bmp, img);
	}
	else
		return -1;

	return 0;
}

CAnimation::~CAnimation()
{
	for (int i = 0; i < anim.size(); i++)
		delete anim[i];
}

/**************************************************************************************/

void CAnimation::computeImage(HBITMAP bmp, CAnimation *image)
{
	CImage *img;
	BITMAP bm;
	HBITMAP bmpOld, hbmOld, hbm;
	HDC hdcFirst, hdcSecond;

	GetObject(bmp, sizeof(BITMAP), &bm);
	has_alpha = (bm.bmBits != 0) && (bm.bmBitsPixel == 32) && !is_opaque;
	image->has_alpha = has_alpha;

	width = image->width;
	height = image->height;
	is_animated = (bm.bmWidth != width) || (bm.bmHeight != height);
	if (!is_animated)
		anim.push_back(img = new CImage(bmp, image));
	else
	{
		if (bm.bmWidth % width != 0)
			width -= bm.bmWidth % width;
		if (bm.bmHeight % height != 0)
			height -= bm.bmHeight % height;

		hdcFirst = CreateCompatibleDC(NULL);
		hdcSecond = CreateCompatibleDC(NULL);

		bmpOld = (HBITMAP)SelectObject(hdcFirst, bmp);

		for (int i = 0; i < bm.bmWidth; i += width)
		{
			//DebugPrint("%d %d %d\n", i, bm.bmWidth, bm.bmHeight);
			hbm = CreateCompatibleBitmap( hdcFirst, width, height );

			hbmOld = (HBITMAP)SelectObject( hdcSecond, hbm );
			BitBlt(hdcSecond, 0, 0, width, height, hdcFirst, i, 0, SRCCOPY);
			SelectObject( hdcSecond, hbmOld );

			anim.push_back(img = new CImage(hbm, image));
		}

		SelectObject(hdcFirst, bmpOld);
		DeleteDC(hdcFirst);
		DeleteDC(hdcSecond);
		DeleteObject(bmp);
	}
	max_size = anim.size() - 1;
	if (!(min || max))
		max = max_size;
	else
	{
		min %= max_size;
		max %= max_size;
		if (min > max)
		{
			int swap = min;
			min = max;
			max = swap;
		}
	}
	current = anim[0];
}

/**************************************************************************************/

void CAnimation::setPos(int &p, bool goon)
{
	if (is_animated)
	{
		if (goon)
		{
			if (p >= max_size)
				pos = loop ? (p % (max_size + 1)) : max_size;
			else
				pos = p;
		}
		else
		{
			if (p >= max)
				pos = loop ? min : max;
			else
				pos = p;
		}
		//DebugPrint("size %d, pos %d, p %d\n", anim.size(), pos, p);
	}
	else
		pos = 0;
	current = anim[pos];
	p = pos;
};

/**************************************************************************************/

CImageStorage::CImageStorage()
{}

CImageStorage::~CImageStorage()
{
	for (int i = 0; i < list.size(); i++)
		delete list[i];
}

/**************************************************************************************/

void CImageStorage::loadAnimation()
{
	char configLine[MAX_LINE_LENGTH],
		 token1[MAX_LINE_LENGTH],
		 token2[MAX_LINE_LENGTH],
		 token3[MAX_LINE_LENGTH],
		 extra[MAX_LINE_LENGTH];
	FILE *f;
	char *tokens[] = {token1, token2, token3};
	CAnimation img;
	CAnimation *anim;
	string name;
	int ret;
	bool is_anim = false;

	f = LCOpen(NULL);
	if (!f)
		return;

	while(LCReadNextConfig(f, "*Layercut", configLine, MAX_LINE_LENGTH))
	{
		ret = LCTokenize (configLine, tokens, 3, extra);
		if (stricmp(token2, "animation") == 0)
		{
			is_anim = true;
			name = string(token3);
		}
		if (is_anim)
		{
			if (stricmp(token2, "image") == 0)
				img.name = string(token3);
			else if (stricmp(token2, "speed") == 0)
				img.speed = atoi(token3);
			else if (stricmp(token2, "between") == 0)
			{
				img.min = atoi(token3);
				img.max = atoi(extra);
			}
			else if (stricmp(token2, "size") == 0)
			{
				if (ret < 3)
					continue;
				img.width = atoi(token3);
				img.height = atoi(extra);
			}
			else if (stricmp(token2, "flags") == 0)
			{
				if (ret < 3)
					continue;
				img.is_opaque	= CharInStr('o', token3) || CharInStr('O', token3);
				img.loop			= CharInStr('l', token3) || CharInStr('L', token3);
				img.finish		= CharInStr('f', token3) || CharInStr('F', token3);
				img.priority	= CharInStr('p', token3) || CharInStr('P', token3);
			}
			else if (stricmp(token2, "~animation") == 0)
			{
				if (!getImage(img.name, 0))
				{
					anim = new CAnimation();
					if (anim->init(img.name, &img))
						delete anim;
					else
					{
						anim->name = name;
						list.push_back(anim);
						qsort(0, list.size() - 1);
					}
				}
				img.reset();
				is_anim = false;
			}
		}
	}
	LCClose(f);
}

/**************************************************************************************/

void CImageStorage::bubble(CAnimation *anim, int deb, int fin)
{
	int ret;
	int milieu;

	while (fin - deb > 1)
	{
		milieu = (deb + fin) / 2;
		ret = list[milieu]->name.compare(anim->name);
		if (ret > 0)
			fin = milieu;
		else if (ret < 0)
			deb = milieu;
		else
			break;
	}
	list.insert(list.begin() + deb, anim);
}

/**************************************************************************************/

void CImageStorage::qsort(int deb, int fin)
{
	int i, j;
	string *milieu;
	CAnimation *temp;

	i = deb;
	j = fin;
	milieu = &(list[(deb + fin) / 2]->name);

	while (i <= j)
	{
		while(list[i]->name.compare(*milieu) < 0) i++;
		while(list[j]->name.compare(*milieu) > 0) j--;

		if(i <= j)
		{
			temp = list[i];
			list[i] = list[j];
			list[j] = temp;
			i++; j--;
		}
	}

	if (i < fin) qsort(i, fin);
	if (deb < j) qsort(deb, j);
}

/**************************************************************************************/

CAnimation *CImageStorage::loadImage(string &str, CAnimation *img)
{
	HBITMAP bmp;
	CAnimation *anim;

	//DebugPrint("begin loading %s\n", str.c_str());

	if (str.compare(".none") == 0)
	{
		//DebugPrint("image .none\n");
		return NULL;
	}
	if ((anim = getImage(str, 0)) != NULL)
	{
		//DebugPrint("image %s already exist\n", str.c_str());
		return anim;
	}

	anim = new CAnimation();
	if (anim->init(str, img))
	{
		//DebugPrint("loading failed %s\n", str.c_str());
		delete anim;
		return NULL;
	}

	list.push_back(anim);
	qsort(0, list.size() - 1);
	//DebugPrint("load %s ok (%d)\n", anim->name.c_str(), list.size());
	//bubble(anim, 0, list.size() - 1);
	return anim;
}

/**************************************************************************************/

CAnimation *CImageStorage::getImage(string &str, int pos)
{
	int milieu = 0;
	int ret = -1;
	int deb = 0, fin = list.size() - 1;

	if (list.empty() || str.empty())
		return NULL;

	do
	{
		milieu = (deb + fin) / 2;
		ret = list[milieu]->name.compare(str);
		if (ret > 0)
			fin = milieu - 1;
		else if (ret < 0)
			deb = milieu + 1;
		else
			break;
	} while (fin - deb >= 0);
	/**/
	/*for (milieu = 0; milieu < list.size(); milieu++)
	{
		ret = list[milieu]->name.compare(str);
		if (ret == 0)
			break;
	}/**/

	if (ret == 0)
	{
		list[milieu]->setPos(pos);
		return (list[milieu]);
	}
	else
		return NULL;
}
